#import "XeePropertiesController.h"
#import "XeeController.h"
#import "XeeImage.h"


@implementation XeePropertiesController

-(void)awakeFromNib
{
	dataarray=nil;

	sectionattributes=[[NSDictionary dictionaryWithObjectsAndKeys:
		[NSFont boldSystemFontOfSize:14],NSFontAttributeName,
	nil] retain];

	NSMutableParagraphStyle *para=[[NSParagraphStyle defaultParagraphStyle] mutableCopy];
	[para setAlignment:NSRightTextAlignment];
	[para setLineBreakMode:NSLineBreakByTruncatingTail];

	labelattributes=[[NSDictionary dictionaryWithObjectsAndKeys:
		[NSFont boldSystemFontOfSize:12],NSFontAttributeName,
		para,NSParagraphStyleAttributeName,
	nil] retain];

	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(frontImageDidChange:)
	name:@"XeeFrontImageDidChangeNotification" object:nil];

	id delegate=[[[NSApplication sharedApplication] keyWindow] delegate];
	if([delegate isKindOfClass:[XeeController class]])
	{
		XeeImage *image=[(XeeController *)delegate image];
		[[NSNotificationCenter defaultCenter] postNotificationName:@"XeeFrontImageDidChangeNotification" object:image];
	}
}

-(void)frontImageDidChange:(NSNotification *)notification
{
	[dataarray release];
	dataarray=[[[notification object] properties] retain];

	[outlineview reloadData];
	[self restoreCollapsedStatusForArray:dataarray];
}

-(void)show
{
	[infopanel orderFront:nil];
}



-(BOOL)outlineView:(NSOutlineView *)view isItemExpandable:(XeePropertyItem *)item
{
	if([item isSubSection]) return YES;
	else return NO;
}

-(int)outlineView:(NSOutlineView *)view numberOfChildrenOfItem:(XeePropertyItem *)item
{
	NSArray *children;
	if(!item) children=dataarray;
	else children=[item value];

	if(!children) return 0;
	return [children count];
}

-(id)outlineView:(NSOutlineView *)view child:(int)index ofItem:(XeePropertyItem *)item
{
	NSArray *children;
	if(!item) children=dataarray;
	else children=[item value];

    return [children objectAtIndex:index];
}

-(id)outlineView:(NSOutlineView *)view objectValueForTableColumn:(NSTableColumn *)col byItem:(XeePropertyItem *)item
{
	NSString *identifier=[col identifier];

	if([item isSubSection])
	{
		if([identifier isEqual:@"label"]) return [[[NSAttributedString alloc]
		initWithString:[item label] attributes:sectionattributes] autorelease];
		else return nil;
	}
	else
	{
		if([identifier isEqual:@"label"]) return [[[NSAttributedString alloc]
		initWithString:[item label] attributes:labelattributes] autorelease];
		// return [item label];
		else return [item value];
	}
}



-(BOOL)outlineView:(NSOutlineView *)view shouldEditTableColumn:(NSTableColumn *)col item:(XeePropertyItem *)item
{
	return NO;
}

-(float)outlineView:(NSOutlineView *)view heightOfRowByItem:(XeePropertyItem *)item
{
	if([item isSubSection]) return 18;
	else return 16;
}

-(void)outlineViewItemDidCollapse:(NSNotification *)notification
{
	XeePropertyItem *item=[[notification userInfo] objectForKey:@"NSObject"];
	NSString *defname=[NSString stringWithFormat:@"propertyListCollapsed.%@",[item label]];
	[[NSUserDefaults standardUserDefaults] setBool:YES forKey:defname];
}

-(void)outlineViewItemDidExpand:(NSNotification *)notification
{
	XeePropertyItem *item=[[notification userInfo] objectForKey:@"NSObject"];
	NSString *defname=[NSString stringWithFormat:@"propertyListCollapsed.%@",[item label]];
	[[NSUserDefaults standardUserDefaults] removeObjectForKey:defname];
}

-(NSString *)outlineView:(NSOutlineView *)view toolTipForCell:(NSCell *)cell rect:(NSRectPointer)rect
tableColumn:(NSTableColumn *)col item:(XeePropertyItem *)item mouseLocation:(NSPoint)mouse
{
	if([item isSubSection]) return nil;
	if([[col identifier] isEqual:@"label"]) return [item label];
	else return [[item value] description];
}


-(void)restoreCollapsedStatusForArray:(NSArray *)array
{
	NSUserDefaults *defaults=[NSUserDefaults standardUserDefaults];
	NSEnumerator *enumerator=[array objectEnumerator];
	XeePropertyItem *item;
	while(item=[enumerator nextObject])
	{
		id value=[item value];
		if([value isKindOfClass:[NSArray class]])
		{
			NSString *defname=[NSString stringWithFormat:@"propertyListCollapsed.%@",[item label]];
			if([defaults boolForKey:defname]) [outlineview collapseItem:item];
			else [outlineview expandItem:item];
			[self restoreCollapsedStatusForArray:value];
		}
	}
}

@end



@implementation XeePropertyOutlineView

-(void)drawRow:(int)row clipRect:(NSRect)clipRect
{
//	int index=[self indexForRow:row];

/*	if(![self isRowSelected:row])
	{
		[[NSColor greenColor] set];
		[self drawRoundedBar:[self rectOfRow:row]];
	}*/

	[super drawRow:row clipRect:clipRect];
}

-(void)drawRoundedBar:(NSRect)rect
{
	NSBezierPath *path=[NSBezierPath bezierPath];
	[path setLineCapStyle:NSRoundLineCapStyle];
	[path setLineWidth:rect.size.height-3];
	float offs=rect.size.height/2.0;
	float x=rect.origin.x+offs+2;
	float y=rect.origin.y+offs;
	[path moveToPoint:NSMakePoint(x,y)];
	[path lineToPoint:NSMakePoint(x+rect.size.width-2*offs-4,y)];
	[path stroke];
}

@end
